#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <bsm/audit.h>
#include <bsm/audit_kevents.h>
#include <bsm/audit_uevents.h>
#include <bsm/audit_record.h>
#include "vio.h"

#define 	BUFLEN	1024
static int fd;
static void *prec;
static int buflen;
static int isOnline; 

UOPEN(int argc, char **argv)
{
    char *option, c;
    int i;
   
   printf("vopen(");
   for (i=0; i<argc; i++)
	printf("%s ", argv[i]);
   printf(")\n");


	/* open native file */

    prec = (void *) malloc(BUFLEN);		/* native record buffer */
    buflen = BUFLEN;
    if (!prec)
 	return VBADOPEN;

   if (argc == 0) {			/* read from stdin off-line */
	fd = 0;
	return 0;
    }
    if ((*argv)[0] == '-') {
		option = *argv;
		while ((c = *++option) != '\0')  {
	    		switch (c) {
				case 'o':
			  		isOnline = 1;
					break;
				default :
			  		printf("vopen: illegal option: %c\n", c);
					exit(1);
	    		}
		}
		argv++;
		argc--;
    }
    if (argc == 1) { 		  
        	fd = open(*argv ,O_RDONLY);
        	if (fd >= 0)  { /* if the open succeeded */
	     		return(OK);
    	        } else {
			free(prec);
	     		return VBADOPEN;
		}
    } else if (argc == 0) {  	
    		fd = 0;   /* read on-line from stdin */
   		return 0;
    } else {
		fprintf(stderr, "vopen: incorrect argument count\n");
		free(prec);
		exit(1);
    }
}
 
 int
 UREAD(char **rec)
 {
 	int rc;
	int tok_id;
	short shtlen; 
	int bytecnt; 
	int prefLen = 4+1+2+2+8;
 	
 	if ((rc = readslow(fd, prec, 1)) == 0)
		return VEOF;
	if (rc < 0)
		return VBADREAD;
	tok_id = *(char *) prec;
    	if (tok_id == AUT_OTHER_FILE) {
	
		/* file token which begins every log file */

 		if ((rc = readslow(fd, (char *) prec+1, 10)) == 0)
			return VEOF;
		if (rc < 10 || rc < 0)
			return VBADREAD;
		memcpy(&shtlen, (char *) prec+1+4+4, 2);
 		if ((rc = readslow(fd, (char *) prec+1+4+4+2, shtlen)) == 0)
			return VEOF;
		if (rc < shtlen || rc < 0)
			return VBADREAD;		
	} else if (tok_id != AUT_HEADER) {
		fprintf(stderr, "expecting a header token while finding %d",
				 tok_id);
		exit(1);
	} else {
		int rest;

 		if ((rc = readslow(fd, (char *) prec+1, prefLen)) == 0)
			return VEOF;
		if (rc<prefLen || rc < 0)
			return VBADREAD;
		memcpy(&bytecnt, (char *) prec+1, 4);
		if (bytecnt > buflen) {
		    do 
			buflen += 1024;
		    while (bytecnt > buflen);
		    if ((prec = realloc(prec, (size_t) buflen)) == 0) {
			fprintf(stderr, "couldn't realloc ADF buffer\n");
			exit(1);
		    }
		}
		rest = bytecnt - (1+prefLen);
 		if ((rc = readslow(fd, (char *) prec+(1+prefLen), rest)) == 0)
			return VEOF;
		if (rc<rest || rc < 0)
			return VBADREAD;
	}
	*rec = (char *) prec;
	return OK;
 }
 
 int
 UCLOSE()
 {
 	int rc;
 	
 	if ((rc = close(fd)) != 0)
 		return VBADCLOSE;
 	return OK;
 }

static int readslow(fd, buf, len) 
int fd;
char* buf;
unsigned len;
{
    int rc, max;

    max = len;
    while (len) {
	rc = read (fd, buf, len);
	if (!isOnline || rc < 0) 
		return rc;
	if (rc == 0)    
		sleep(1);
	if (rc == len)  
		return max;
	buf += rc;
	len -= rc;
    }
    return 0;
}

