/*
 :******************************************************************
 :* MODULE-DESCRIPTION
 :*
 :* TITLE         (/ Parser of a Library Description file /)
 :*
 :* NAME          parseLib.c
 :*
 :* COPYRIGHT     (C)   SIEMENS NIXDORF SOFTWARE SA.  1993
 :*                     ALL RIGHTS RESERVED
 :*
 :* MEMORY-CLASS  6
 :*
 :* PROG-CONVENTS (/ <internal conventions (instance, names,
 :*                  external data, ...). None -> empty text>     /)
 :*
 :* PROCEDURE     main
 :* REMARKS       (/ parse a library description file and generate
 :*                  a make file and two headrer files necessary for
 :*                  generating a new ASAX executable /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     parseLDF
 :* REMARKS       (/ parser of a library description file         /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     parseLDFile
 :* REMARKS       (/ parser of Library Description file         /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     description
 :* REMARKS       (/ parsing a descriptor of a predefined routines /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     head
 :* REMARKS       (/ parsing the type and name of a predefined routines /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     param_list
 :* REMARKS       (/ parsing the list of parameter descriptors including
 :*                  the parenthesis.                           /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     paramList
 :* REMARKS       (/ parsing the list of parameter descriptors excluding
 :*                  the parenthesis.                           /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     variable_arity
 :* REMARKS       (/ parsing the (common) descriptor of a varible arity
 :*                  predefined routine                         /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     paramList1
 :* REMARKS       (/ parsing the list of parameter descriptors for a
 :*                  fixed arity predefined routine               /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     param
 :* REMARKS       (/ parsing a parameter descriptor               /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     install_descr
 :* REMARKS       (/ adds a parameter descriptor to the table of parameter
 :*                  descriptors                                  /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     linkLib
 :* REMARKS       (/ generate a make file for ASAX                /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     dparse_LDF
 :* REMARKS       (/ generate a C-initialization file and prototype file /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     mergeLib
 :* REMARKS       (/ merging of descriptor tables related to the library /)
 :* ENTRY         EXTERNAL
 :*
 :*-----------------------------------------------------------------
 :*
 :* VERSION       002
 :*   CRDATE      1993-07-06
 :*   AUTHOR      (/ Mounji Abdelaziz    , FUNDP /)
 :*   UPDATE      (/ original issue /)
 :*
 :* END-MODULE-DESCRIPTION parseLib.c
 :******************************************************************
*/
#include <stdio.h>
#include <ctype.h>
#include "specTab.h"

  /* initialization of strings to be written to an ASAX make file   */

char *apath = "APATH=";
char *lib_files= "usr_lib=";
char *obj_file = "obj_file=";
char *mode_file= "mode_file=standard.o";
char *asaxExec = "asaxExec=asax.exec";
char
*cc1 ="$APATH/asaxKernel.o .__std.o $APATH/$mode_file $usr_lib $obj_file -o $asaxExec";

    /* table of type names supported by  a library description file */

char *typesTab[4] = {"integer", "bytestring", "fieldstring", "undef"};
char *passtypeTab[2] = {"value", "ref"};

int LineNbre   = 1;    /* the current line during parsing              */
int ErrorFound = 0;    /* = 1 if an error of some sort was encountered */


int Stop1[lgStop1] = {nosymbx};       /* end of file */

	/* initialization of internal tables length */

int sys_arg_lg = 0;
int sys_fp_lg = 0;
int user_arg_lg = 0;
int user_fp_lg = 0;
int tObjMod_lg = 0;
int tObjSysMod_lg = 0;    /* length of tObjMod[] after parsing sys_ldf    */

fp_descr sys_fp[MaxFp];   /* table of predefined routine descriptors      */
arg_descr sys_arg[MaxFpArg];/* table of argument descriptors                */

char *tObjMod[MaxObjMod]; /* an entry in this table represents the name   */
			  /* of the object module implementing a          */
			  /* predefined routine                           */

FILE *rfp;                /* file descriptor of library description file  */

char *CurrMod_name = "sys_ldf";  /* internal name for the file containing */
				 /* the description of the predefined     */
				 /* routines                              */

char *defaultName = "ldf";       /* default name for library description  */
				 /* file for user-defined routines        */

/*
 :***********************************************************************
 :* UNIT-BODY parseLib.main
 :*
 :* TITLE         (/ parser of a library description file           /)
 :*
 :* PURPOSE       (/ parse a library description file and generate
 :*                  a make file and two header files necessary for
 :*                  generating a new ASAX executable               /)
 :*
 :*
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-06
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                     - CurrMod_name       : name of the user library
 :*                                            description file
 :*
 :*                     - defaultName        : default name for a user
 :*                                            library description file
 :*
 :*                     - fp_descr sys_fp[]  : table of predefined routine
 :*                                            descriptors
 :*
 :*                     - arg_descr sys_arg[]: table of argument descriptors
 :*
 :*                     - tObjMod[MaxObjMod] : an entry in this table
 :*                                            represents the name of the
 :*                                            of the object module
 :*                                            implementing a predefined
 :*                                            routine
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator
 :*                               ASAX: Specification of
 :*                          some extensions to the analyzer           /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/
 :*             - 'argv[]' is an array of 'argc' strings
 :*
 :*                                                                      /)
 :*                                                             END-Pre
 :*   Post   (/
 :*      I/ If
 :*               1)   - argc = 3
 :*                    - argv[1] = "-s"
 :*                    - argv[2] = a non null string say 'prefix'
 :*                    - the name 'prefix' with the extension '.ldf' is a
 :*                      valid file name
 :*                    - the string CurrMod_name is a valid file name
 :*          OR
 :*               2)   - argc = 2
 :*                    - argv[1] = a non null string say 'prefix'
 :*                    - the name 'prefix' with the extension '.ldf' is a
 :*                      valid file name
 :*                    - the string CurrMod_name is a valid file name
 :*
 :*          OR
 :*               3)   - argc = 1
 :*                    - the name "ldf" (default name) is a valid file name
 :*                    - the string CurrMod_name is a valid file name
 :*
 :*          OR
 :*               4)   - argc = 2
 :*                    - argv[1] = "-s"
 :*                    - the name "ldf" (default name) is a valid file name
 :*                    - the string CurrMod_name is a valid file name
 :*
 :*        Then
 :*          in absence of syntactic or semantic errors in either of
 :*          the two above files, three text files are generated. The
 :*          content and naming conventions of these files are specified
 :*          respectivelly in the following:
 :*
 :*          NAMING CONVENTIONS:
 :*             in case 1) and 2) above:
 :*
 :*                  - 'prefix'.makeAsax
 :*                  - 'prefic'.asax_lib.h
 :*                  - 'prefix'.proto.h
 :*
 :*             in case 3) and 4)
 :*
 :*                  - makeAsax
 :*                  - asax_lib.h
 :*                  - proto.h
 :*
 :*          CONTENT:
 :*           * The make file: (mode 777)
 :*                Let module_1, ..., module_k (k>=0) be the names of the
 :*                object modules listed in the two input files. Then the
 :*                make file (makeAsax or 'prefix'.makeAsax) have the
 :*                following content:
 :*
 :*                lib_files='module_1.o module2.o ...  module_k.o'
 :*                obj_file=obj_file_name
 :*                mode_file=standard.o
 :*                asaxExec=asax.exec
 :*
 :*
 :*                cc -c std.c std.o
 :*
 :*                case $1 in
 :*                        0) obj=NADF_file.o;;
 :*                        *) obj="$obj_file NADF_file.o"
 :*                esac
 :*
 :*
 :*                cc $mode_file emulate.o $obj  \
 :*                stack.o parse.o usage.o globalVar.o rule_decl.o action.o \
 :*                guardAct.o condition.o trigger.o arglist.o term.o  \
 :*                expr.o string.o printbox.o std.o recov.o rdescr.o \
 :*                error.o merge.o $lib_files scan.o  -o $asaxExec
 :*
 :*            * asax_lib.h (or 'prefix'.asax_lib.h):   (mode 777)
 :*              contains the C-declaration (initialization) of the tables
 :*              'tfp_descr[]' and 'targ_descr[]' with respect to description
 :*              of these routines listed in the two input files.
 :*
 :*            * proto.h (or 'prefix'.proto.h):         (mode 777)
 :*              contains the C-protoypes (old style C) for the routines
 :*              listed in the two input files.
 :*
 :*      II/ If
 :*               1)   - argc = 2
 :*                    - argv[1] = "-r"
 :*                    - the string CurrMod_name is a valid file name
 :*            OR
 :*
 :*               2)   - argc = 3
 :*                    - argv[1] = "-r"
 :*                    - argv[2] = a non null string say 'prefix'
 :*                    - the string CurrMod_name is a valid file name
 :*        Then
 :*          in absence of syntactic or semantic errors in the file named
 :*          'CurrMod_name', three text files are generated. The content
 :*          content and naming conventions of these files are specified
 :*          respectivelly in the following:
 :*
 :*          NAMING CONVENTIONS:
 :*             in case 1) above:
 :*
 :*                  - 'prefix'.makeAsax
 :*                  - 'prefic'.asax_lib.h
 :*                  - 'prefix'.proto.h
 :*
 :*             in case 2)
 :*
 :*                  - makeAsax
 :*                  - asax_lib.h
 :*                  - proto.h
 :*
 :*          CONTENT:
 :*              the content of these three output files is the one obtained
 :*              by from an empty file 'ldf' (or 'prefix'.ldf) in the first
 :*              case.
 :*
 :*      III/   Otherwise, a list of errors is reported on stdout         /)
 :*                                                             END-Post
 :*   RETURNS NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*
 :*   REMARKS          (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parseLib.main
 :******************************************************************
*/

int main(argc, argv)
int argc;                         /* argument counter                 */
char *argv[];                     /* argument vector                  */
{
    int mode;                     /* reset mode or system mode        */
    char c;                       /* ='r' or 's' reset or system mode */
    char *option;                 /* the option                       */
    char *file;                   /* input file name                  */
    char *descr, *proto, *make;   /* the three output file names      */
    char  *strcat();
    FILE *fd1;                    /* table initialization file        */
    FILE *fd2;                    /* prototypes file                  */
    char buf[2000];
    char *bufp = buf;

    if (argc > 3) {
	error(fatal, 12);      /* too many arguments               */
	exit(1);
    }
    argv++;
    argc--;
    if (argc >= 1) {
        if (**argv == '-') {
	    option = *argv++;
	    argc--;
	    while ((c = *++option) != '\0')  {
	       switch (c) {
		   case 's':
			  mode = system;           /* system mode */
			  break;
		   case 'r':
			  mode = reset;            /* reset mode  */
			  break;
		   default :
			  error1(fatal, 8, option);  /* illegal option */
			  exit(1);
	       }
	    }
        }
    }
	/* open predefined routines library description file */
    sprintf(bufp, "%s/%s", ASAXPATH, CurrMod_name);
    if ((rfp = fopen(bufp, "r")) == NULL) {
	error1(fatal, 6, bufp);
	exit(1);
    }
	/* parse this file and generate internal tables */

    parse_LDF(rfp, sys_fp, &sys_fp_lg, sys_arg, &sys_arg_lg);
    fclose(rfp);

    if (ErrorFound)   /* terminate in case of errors in the first file */
	exit(1);
    LineNbre = 1;
    if (argc >= 1)
        CurrMod_name = *argv;      /* get the file name from command line */
    else
        CurrMod_name = "";  
		   /* allocate memory for filenames */

    if ((file = (char *) malloc(60)) == NULL) {
	error(fatal, 9);           /* allocation problem */
	exit(1);
    }
    if ((descr = (char *) malloc(60)) == NULL) {
	error(fatal, 9);           /* allocation problem */
	exit(1);
    }
    if ((proto = (char *) malloc(60)) == NULL) {
	error(fatal, 9);           /* allocation problem */
	exit(1);
    }
    if ((make = (char *) malloc(60)) == NULL) {
	error(fatal, 9);           /* allocation problem */
	exit(1);
    }

	/* build the input and output file names according to naming */
	/* conventions */

    if (strcmp(CurrMod_name, "")) {     /* if a prefix was given */
	strcpy(file, CurrMod_name);
	file = strcat(file, ".ldf");    /* add library name extension */
	strcpy(descr, *argv);
	strcpy(proto, *argv);
	strcpy(make, *argv);
	descr = strcat(descr, ".asax_lib.h");
	proto = strcat(proto, ".proto.h");
	make = strcat(make, ".makeAsax");
    } else {                            /* default case */
	strcpy(file, defaultName);
	strcpy(descr, "asax_lib.h");
	strcpy(proto, "proto.h");
	strcpy(make, "makeAsax");
    }
    CurrMod_name = file;     /* name of the user library descritpion file */
    tObjSysMod_lg = tObjMod_lg; /* save the current length */
    if (mode != reset) {

	/* open user-defined routines library description file */

	if ((rfp = fopen(file, "r")) == NULL) {
	    error1(fatal, 6, file);
	    exit(1);
	}

	/* parse this file and generate internal tables */
	parse_LDF(rfp, user_fp, &user_fp_lg, user_arg, &user_arg_lg);
	fclose(rfp);
	if (ErrorFound)  /* terminate in case of errors */
	    exit(1);
    }
      /* open output files */

    if ((fd1 = fopen(proto, "w")) == NULL) {   /* prototypes file */
	error1(fatal, 6, proto);
	exit(1);
    }
    if ((fd2 = fopen(descr, "w")) == NULL) {   /* C-initialization file */
	error1(fatal, 6, descr);
	exit(1);
    }

    dparse_Ldf(fd1, fd2);         /* generate the two files */
    fclose(fd1);
    fclose(fd2);                  /* close the two files    */
    linkLib(make);                /* generate the make file */

		 /* release memory areas for file names */

    free(descr);
    free(proto);
    free(make);
    return(0);
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.parseLDF
 :*
 :* TITLE         (/ parser of the Library Description file /)
 :*
 :* PURPOSE       (/ this function receives a file pointer of a Library
 :*                  Description file open in "READ" mode. In absence of
 :*                  errors, it initializes the table 'us_fp[]' and 'us_arg[]'
 :*                  with routine and argument descriptors contained in this
 :*                  file /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-07
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*
 :*                     - tObjMod[MaxObjMod] : an entry in this table
 :*                                            represents the name of the
 :*                                            of the object module
 :*                                            implementing a predefined
 :*                                            routine
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator
 :*
 :*                               ASAX: Specification of
 :*                          some extensions to the analyzer           /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - 'fp' is the file pointer of the library Description file
 :*               open in "READ" mode
 :*             - 'us_fp[]'  is an array of predefined routine descriptors;
 :*             - us_fp_lg   is a pointer to the length of 'us_fp[]'
 :*             - 'us_arg[]' is an array of argument descriptors;
 :*             - us_arg_lg  is a pointer to the length of 'us_arg[]'  /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - if the content of the file 'fp' is:  (n>=0, m>=0)
 :*
 :*      ret_typ_1 routine_name_1 (t1_1:mode1_1, ..., t1_n1:mode1_n1): obj_1;
 :*                                     .
 :*                                     .
 :*                                     .
 :*      ret_typ_m routine_name_m (t1_m:modem_1, ..., tm_nm:modem_nm): obj_m;
 :*
 :*      i.e. the file is corecte with respect to the library description
 :*      file syntax.
 :*
 :*      Then
 :*         * the table 'us_fp[]' contains 'm' entries defined as follows:
 :*
 :*             {ROUTINE_NAME_i, ROUTINE_PTR_i, N_i, RET_TYP_i, K_i}
 :*
 :*           where
 :*             - ROUTINE_NAME_i is a pointer to the function routine_name_i;
 :*             - 'N_i' = 'n_i' if routine_name_i is of a fixed arity
 :*                     = -1    if it is of a variable number of arguments;
 :*             - RET_TYP_i = integer    if ret_typ_i = "integer"
 :*                         = bytestring if ret_typ_i = "string"
 :*                         = undef      if ret_typ_i is empty
 :*             - K_i    if routine_name_i is of a fixed arity,
 :*                      is an index in the table 'us_arg[]' corresponding to
 :*                      argument descriptors of routine_name_i
 :*                      if it is of a variable number of arguments,
 :*                      is the passing mode ('ref' or 'value' common to all
 :*                      arguments;
 :*
 :*           'us_fp[]' is sorted in alphabetic order on 'routine_name_i's
 :*
 :*         * the table 'us_arg[]' contains (n1 + ... + nm) entries defined
 :*           as follows:
 :*               for each entry 'i' in the table 'us_fp[]' it corresponds
 :*               'ni' entries in 'us_arg[]' starting from index 'K_i' such
 :*               that:
 :*                     us_arg[j].tpass = 'value'      if 'modei_j' = val
 :*                     us_arg[j].tpass = 'ref'        if 'modei_j' = ref
 :*                     us_arg[j].targ  = 'integer'    if 'ti_j'    = integer
 :*                     us_arg[j].targ  = 'bytestring' if 'ti_j'    = string
 :*
 :*         * (*us_fp_lg  = m)
 :*         * (*us_arg_lg = n1 + ... + nm)
 :*
 :*      - otherwise one or more errors is output onb stdout          /)
 :*
 :*                                                             END-Post
 :*   RETURNS NONE
 :*
 :*   REMARKS          (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parseLib.parseLDF
 :******************************************************************
*/
parse_LDF(fp, us_fp, us_fp_lg, us_arg, us_arg_lg)
FILE *fp;                     /* file pointer to be parsed     */
fp_descr *us_fp;              /* table of routine descriptors  */
arg_descr *us_arg;            /* table of argument descriptors */
int *us_fp_lg, *us_arg_lg;    /* pointer to table lengths      */
{
    int S1[Mlg], lg1, S2[Mlg], lg2, S3[Mlg], lg3; /* auxiliary staops   */
                                   /* memory allocation */

    /* allocate namle for identifiers */

    if ((name = (char *) malloc(ident_lg)) == NULL) {
	error(fatal, 9);           /* allocation problem */
	exit(1);
    }
    get_char(fp);                      /* get the first character */
    getsymb(fp);                       /* get the first token     */

    /* build stops */

    insert_stop(identifierx, Stop1, lgStop1, S3, &lg3);
    insert_stop(intx, S3, lg3, S2, &lg2);
    insert_stop(strx, S2, lg2, S1, &lg1);
    SyntaxCheck(S1, lg1);
    if (symb.ibsymb != nosymbx) {      /* if end of file reached */
	parseLDFile(us_fp, us_fp_lg, us_arg, us_arg_lg, Stop1, lgStop1);
    }
    free(name);                        /* release memory         */
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.parseLDFile
 :*
 :* TITLE         (/ parsing of a library description file   /)
 :*
 :* PURPOSE       (/ parse a sequence of symbols representing a library
 :*                  description file and generate the internal tables
 :*                  'us_fp[]' and 'usarg[]' accordingly if no
 :*                  error is found or recover from this error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*
 :*       INVARIANT:
 :*
 :*      <descriptor_1><;>...<descriptor_i><;>... <descriptor_n><.>
 :*      <--------------------------------><------------------->
 :*               treated  (=alpha)                  to be treated
 :*
 :*
 :*               where
 :*
 :*                  <descriptor_i> is a valid routine descriptor i=1..N
 :*                  - the CSS is <;>... <descriptor_n><.>
 :*
 :*        the tables 'us_fp[]' and 'us_arg[]' are initailized according to
 :*        the POST condition applied to <descriptor_1>, ..., <descriptor_i>
 :*
 :*
 :*                                                               /)
 :*   DATA       (/
 :*                       - symb       : holds the internal representation
 :*                                      of the last read symbol.
 :*                       - nextchar   : the last character being read.
 :*                       - ErrorFound : = TRUE if an error was found
 :*                                      = FALSE otherwise
 :*
 :*                       - tObjMod[MaxObjMod] : an entry in this table
 :*                                            represents the name of the
 :*                                            of the object module
 :*                                            implementing a predefined
 :*                                            routine
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator
 :*
 :*                               ASAX: Specification of
 :*                          some extensions to the analyzer           /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - us_fp[]    is an array of predefined routine descriptors;
 :*             - us_fp_lg   is a pointer to the length of us_fp[]
 :*             - us_arg[]   is an array of argument descriptors;
 :*             - us_arg_lg  is a pointer to the length of us_arg[]
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)
 :*
 :*                                                             END-Pre
 :*
 :*   Post   (/ - if the CSS is:  (n>=0, m>=0)
 :*
 :*      ret_typ_1 routine_name_1 (t1_1:mode1_1, ..., t1_n1:mode1_n1): obj_1;
 :*                                     .
 :*                                     .
 :*                                     .
 :*      ret_typ_m routine_name_m (t1_m:modem_1, ..., tm_nm:modem_nm): obj_m;
 :*
 :*      i.e. the file is corecte with respect to the library description
 :*      file syntax.
 :*
 :*      Then
 :*         * the table 'us_fp[]' contains 'm' entries defined as follows:
 :*
 :*             {ROUTINE_NAME_i, ROUTINE_PTR_i, N_i, RET_TYP_i, K_i}
 :*
 :*           where
 :*             - ROUTINE_NAME_i is a pointer to the function routine_name_i;
 :*             - 'N_i' = 'n_i' if routine_name_i is of a fixed arity
 :*                     = -1    if it is of a variable number of arguments;
 :*             - RET_TYP_i = integer    if ret_typ_i = "integer"
 :*                         = bytestring if ret_typ_i = "string"
 :*                         = undef      if ret_typ_i is empty
 :*             - K_i    if routine_name_i is of a fixed arity,
 :*                      is an index in the table 'us_arg[]' corresponding to
 :*                      argument descriptors of routine_name_i
 :*                      if it is of a variable number of arguments,
 :*                      is the passing mode ('ref' or 'value' common to all
 :*                      arguments;
 :*
 :*           'us_fp[]' is sorted in alphabetic order on 'routine_name_i's
 :*
 :*         * the table 'us_arg[]' contains (n1 + ... + nm) entries defined
 :*           as follows:
 :*               for each entry 'i' in the table 'us_fp[]' it corresponds
 :*               'ni' entries in 'us_arg[]' starting from index 'K_i' such
 :*               that:
 :*                     us_arg[j].tpass = 'value'      if 'modei_j' = val
 :*                     us_arg[j].tpass = 'ref'        if 'modei_j' = ref
 :*                     us_arg[j].targ  = 'integer'    if 'ti_j'    = integer
 :*                     us_arg[j].targ  = 'bytestring' if 'ti_j'    = string
 :*
 :*         * (*us_fp_lg  = m)
 :*         * (*us_arg_lg = n1 + ... + nm)
 :*         * the CSS is <nosymb>
 :*
 :*      - otherwise one or more errors is output on stdout          /)
 :*
 :*                                                             END-Post
 :*   RETURNS NONE
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.
 :*                                                         /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parseLib.parseLDFile
 :******************************************************************
*/
parseLDFile(us_fp, us_fp_lg, us_arg, us_arg_lg, Stop, lg)
fp_descr *us_fp;              /* table of routine descriptors  */
arg_descr *us_arg;            /* table of argument descriptors */
int *us_fp_lg, *us_arg_lg;    /* pointer to table lengths      */
int Stop[], lg;               /* The stops */
{
    int S1[Mlg], lg1, S2[Mlg], lg2, S3[Mlg], lg3;  /* auxiliary stops */

    /* build stops */
    insert_stop(semicolon, Stop, lg, S3, &lg3);
    insert_stop(period, S3, lg3, S1, &lg1);
    insert_stop(strx, S1, lg1, S2, &lg2);
    insert_stop(intx, S2, lg2, S3, &lg3);
    insert_stop(identifierx, S3, lg3, S2, &lg2);

    /* parse first descriptor */

    descriptor(us_fp, us_fp_lg, us_arg, us_arg_lg, S1, lg1);
    while (symb.ibsymb == semicolon) {  /* loop for other descriptors */
	Expect(semicolon, S2, lg2);     /* get  semicolon */
    /* parse next descriptor */
	descriptor(us_fp, us_fp_lg, us_arg, us_arg_lg, S1, lg1);
    }
    Expect(period, Stop, lg);     /* get ending period */
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.descriptor
 :*
 :* TITLE         (/ parsing of a library description file   /)
 :*
 :* PURPOSE       (/ parse a sequence of symbols representing a library
 :*                  description file and generate the internal tables
 :*                  'us_fp[]' and 'usarg[]' accordingly if no
 :*                  error is found or recover from this error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                       - symb       : holds the internal representation
 :*                                      of the last read symbol.
 :*                       - nextchar   : the last character being read.
 :*                       - ErrorFound : = TRUE if an error was found
 :*                                      = FALSE otherwise
 :*
 :*                       - tObjMod[MaxObjMod] : an entry in this table
 :*                                            represents the name of the
 :*                                            of the object module
 :*                                            implementing a predefined
 :*                                            routine
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator
 :*
 :*                               ASAX: Specification of
 :*                          some extensions to the analyzer           /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - us_fp[]    is an array of predefined routine descriptors;
 :*             - us_fp_lg   is a pointer to the length of us_fp[]
 :*             - us_arg[]   is an array of argument descriptors;
 :*             - us_arg_lg  is a pointer to the length of us_arg[]
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)
 :*
 :*                                                             END-Pre
 :*
 :*   Post   (/   if the CSS is of the following form:
 :*                 <descriptor>";"<rest>
 :*               where
 :*                  - <descriptor> is a sequence of symbols representing
 :*                                      a valid routine descriptor i=1..N
 :*                  - <rest> is any character sequence
 :*               then the effect is the following:
 :*                * a new entry corresponding to <descriptor> is added
 :*                  to 'us_fp[]' and as many entries are added to 'us_arg[]'
 :*                  as there is parameter descriptors in <descriptor>;
 :*                * let S0 be the set of entries in the table 'tObjMod[]',
 :*                  lg_0 its length and 'ObjModule' the name of the object
 :*                  module in <descriptor>. Then the set of entries in
 :*                  'tObjMod[]' is S1 = S0 U {'ObjModule'}, tObjMod_lg =#S1;
 :*                * The CSS is ";"<rest>.
 :*                otherwise
 :*                  - the CSS is of the form:
 :*                         <first_Stop><rest>
 :*                    where <first_Stop> is the first symbol in
 :*                                 s1 s2 ... sn
 :*                          such that <first_Stop> is in 'Stop[]'
 :*                  - one or more error messages are output to stdout
 :*                  - 'ErrorFound' is set to TRUE           /)
 :*
 :*                                                             END-Post
 :*   RETURNS NONE
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.
 :*                                                         /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parseLib.descriptor
 :******************************************************************
*/
descriptor(us_fp, us_fp_lg, us_arg, us_arg_lg, Stop, lg)
fp_descr *us_fp;              /* table of routine descriptors  */
arg_descr *us_arg;            /* table of argument descriptors */
int *us_fp_lg, *us_arg_lg;    /* pointer to table lengths      */
int Stop[], lg;               /* The stops */
{
    int S1[Mlg], lg1, S2[Mlg], lg2, S3[Mlg], lg3;  /* auxiliary stops */
    int i;                    /* index of the descriptor in 'us_fp[]' */

    /* build stops */

    insert_stop(identifierx, Stop, lg, S3, &lg3);
    insert_stop(colon, S3, lg3, S2, &lg2);
    insert_stop(lparenthesis, S2, lg2, S1, &lg1);

 /* parse the returned type and function name then install its descriptor */

    i = head(us_fp, us_fp_lg, us_arg_lg, S1, lg1);

 /* parse the parameter list then install its descriptors in 'us_arg[]'  */

    param_list(i, us_fp, us_arg, us_arg_lg, S2, lg2);
    Expect(colon, S3, lg3);                  /* read the colon           */
    ExpectName(name, Stop, lg);              /* read object module name  */
    add_obj_mod(name);                       /* install it in tObjMod[]' */
}

/*
 :******************************************************************
 :* UNIT-BODY parseLib.add_obj_mod
 :*
 :* TITLE         (/ install a new field name  /)
 :*
 :* PURPOSE       (/ add a new entry to the table mapping field names to
 :*                  their audit data id /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-12-02
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                     - tObjMod[MaxObjMod] : an entry in this table
 :*                                            represents the name of the
 :*                                            of the object module
 :*                                            implementing a predefined
 :*                                            routine
 :*                     - tModObj_lg         : current length of 'tObjMod[]'
 :*                     - MaxObjMod          : maximum length of 'tObjMod[]'
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - 'name' is a string representing a field name
 :*             - the set of entries in 'tObjMod[]' is:
 :*               S = {name_0, ..., name_n}  (n=tObjMod_lg - 1)
 :*             - 'tObjMod_lg' = lg_0
 :*                                                                     /)
 :*                                                             END-Pre
 :*
 :*   Post  (/ - if 'tObjMod_lg'<'MaxObjMod'-1 and 'name' is not in S:
 :*                 - 'tObjMod_lg' = lg_0 + 1
 :*                 - the set of entries in 'tObjMod[]' is:
 :*                              S U {name}
 :*            - if 'tObjMod_lg'<'MaxObjMod'-1 and 'name' is in S:
 :*                 the following error message:
 :*                 "already part of the library: routine_name"
 :*                 is output on stdout
 :*            - otherwise if 'tObjMod_lg'>='MaxObjMod'-1 the error message:
 :*                       "table overflow!: tObjMod"
 :*               is output on stdout
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS          (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.add_obj_mod
 :******************************************************************
*/
add_obj_mod(name) /* adds a new entry in the object modules table */
char *name;
{
    int i;     /* loop counter */
    int found; /* = TRUE if name is in name_0, ..., name_i-1 */
    char *s;

    i=0;
    found = FALSE;

	 /* search module name */

    while (!found && i < tObjMod_lg) {
	if (strcmp(tObjMod[i++], name) == 0)
	    found = TRUE;
    }
    if (!found) {
	if (tObjMod_lg >= MaxObjMod) {/* no overflow of tObjMod table */
	    error1(fatal, 11, "tObjMod");
	    exit(1);
	}
	if ((s = (char *) malloc(strlen(name) + 1)) == NULL) {
	    error(fatal, 9);               /* allocation problem  */
	    exit(1);
	}
	strcpy(s, name);                      /* copy module name    */
	tObjMod[tObjMod_lg++] = s;            /* install module name */
    }
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.head
 :*
 :* TITLE         (/ parsing the type and name of a predefined routines /)
 :*
 :* PURPOSE       (/ parses the returned type and name of a predefined
 :*                  routine and install it in us_fp[] if no error is found
 :*                  or recover from this error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-12-02
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - us_fp[]    is an array of predefined routine descriptors;
 :*             - us_fp_lg   is a pointer to the length of us_fp[]  (= lg0)
 :*             - us_arg_lg  is a pointer to the length of us_arg[]
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)  /)
 :*
 :*                                                             END-Pre
 :*
 :*   Post  (/ if
 :*              - the CSS if of the form:
 :*
 :*                 <ret_typ><routine_name>"("<rest>
 :*
 :*                where  <ret_typ> is    empty
 :*                                or  "integer"
 :*                                or  "string"
 :*                       <routine_name> is a string representing a routine
 :*                                    name
 :*                       <rest> is any character sequence
 :*              - usfp[k].name <> routine_name (0<= k < *us_fp_lg)
 :*              - *us_fp_lg < Maxfp
 :*
 :*            then
 :*              - us_fp[i].name = routine_name
 :*              - us_fp[i]argp = *us_arg_lg
 :*                where i is such that:
 :*                     - 0 <= i <lg0 + 1
 :*                     - us_fp[] is alphabetically sorted on routine names
 :*              - *us_fp_lg = lg0 + 1
 :*              - the CSS is "("<rest>
 :*            otherwise
 :*              - the CSS is of the form:
 :*                     <first_Stop><rest>
 :*                where <first_Stop> is the first symbol in
 :*                             s1 s2 ... sN
 :*                      such that <first_Stop> in 'Stop[]'
 :*              - one or more error messages are output to stdout
 :*              - 'ErrorFound' is set to TRUE
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  "int" (/ -1 in case of error or an index in 'us_fp[]' /)
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.                             /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.head
 :******************************************************************
*/
head(us_fp, us_fp_lg, us_arg_lg, Stop, lg)
fp_descr *us_fp;            /* table of predefined routine descriptors */
int *us_fp_lg, *us_arg_lg;  /* table lengths */
int Stop[], lg;             /* the stops */
{
    int S1[Mlg], lg1;       /* auxiliary stops */
    int i;                  /* index in us_fp[] of this descriptor */

    /* build stops */
    insert_stop(identifierx, Stop, lg, S1, &lg1);
    switch (symb.ibsymb) {

	case identifierx : /* no returned type */
			   ExpectName(name, Stop, lg);
			   i= install_descr(us_fp, us_fp_lg, us_arg_lg,name);
			   us_fp[i].rettype = undef; /* fieldstring; */
			   break;
	case intx        : /* returned type is integer */
			   Expect(intx, S1, lg1);
			   ExpectName(name, Stop, lg);
			   i= install_descr(us_fp, us_fp_lg, us_arg_lg,name);
			   us_fp[i].rettype = integer;
			   break;
	case strx        : /* returned type is string */
			   Expect(strx, S1, lg1);
			   ExpectName(name, Stop, lg);
			   i= install_descr(us_fp, us_fp_lg, us_arg_lg,name);
			   us_fp[i].rettype = bytestring;
			   break;
	default          :
			   error(syntax, 1);
    }                      /* type or routine name expected */
    if (ErrorFound)
	return (-1);      /* an error was found */
    else
	return (i);       /* index in us_fp of the installed descriptor */
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.param_list
 :*
 :* TITLE         (/ parsing the list of parameter descriptors including
 :*                  the parenthesis                                    /)
 :* PURPOSE       (/ parses a parameter descriptor list of a predefined
 :*                  routine and install their descriptors in us_arg[] if no
 :*                  error is found or recover from this error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-08
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - i          is an index of the routine descriptor in us_fp[]
 :*             - us_fp[]    is an array of predefined routine descriptors;
 :*             - us_arg[]   is an array of routine argument descriptors;
 :*             - us_arg_lg  is a pointer to the length of us_arg[]
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)  /)
 :*
 :*                                                             END-Pre
 :*
 :*   Post  (/ if
 :*              - the CSS if of the form:
 :*
 :*                 "("<paramList>")" <rest>
 :*
 :*                where  <paramList> is a valid parameter descriptor list
 :*                       <rest> is any character sequence
 :*              - us_fp[k].name <> routine_name (0<= k < *us_fp_lg)
 :*              - *us_fp_lg < Maxfp
 :*
 :*            then
 :*             1) if <paramList> is empty
 :*                then nothing is done
 :*             2) if <paramList> is of the form:
 :*                         "undef"<passing_mode>
 :*                  where <passing_mode> is "ref" or "val"
 :*                then
 :*                  - us_fp[i].arity = -1
 :*                  - us_fp[i].argp = ref   if <passing_mode> = "ref"
 :*                                  = value if <passing_mode> = "val"
 :*             3) if <paramList> is of the form: (n>=1)
 :*                   <type_1>":"<mode_1>"," ... ","<type_n>":"<mode_n>
 :*                  where <type_i> = "integer" or "string"
 :*                        <mode_i> = "ref" or "val"
 :*                  - us_fp[i].arity = n
 :*                  - for (j = 0, ..., n)
 :*                    us_arg[lg_0 + j].tpass = ref   if <mode_j> = "ref"
 :*                                           = value if <mode_j> = "val"
 :*                    us_arg[lg_0 + j].targ
 :*                                 = integer    if <type_j> = "integer"
 :*                                 = bytestring if <type_j> = "string"
 :*
 :*            in all cases, the CSS is <rest>
 :*          Otherwise
 :*              - the CSS is of the form:
 :*                     <first_Stop><rest>
 :*                where <first_Stop> is the first symbol in
 :*                             s1 s2 ... sN
 :*                      such that <first_Stop> in 'Stop[]'
 :*              - one or more error messages are output to stdout
 :*              - 'ErrorFound' is set to TRUE
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.                             /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.param_list
 :******************************************************************
*/
param_list(i, us_fp, us_arg, us_arg_lg, Stop, lg)
int i;                      /* index of the descr. in us_fp[]          */
fp_descr *us_fp;            /* table of predefined routine descriptors */
arg_descr *us_arg;          /* table of routine argument descriptors   */
int *us_arg_lg;             /* pointer to the length of us_arg[]       */
int Stop[], lg;             /* stops */
{
    int S1[Mlg], lg1, S2[Mlg], lg2, S3[Mlg], lg3; /* auxiliary stops */

   /* build stops */

    insert_stop(rparenthesis, Stop, lg, S2, &lg2);
    insert_stop(undefx, S2, lg2, S1, &lg1);
    insert_stop(intx, S1, lg1, S3, &lg3);
    insert_stop(strx, S3, lg3, S1, &lg1);
    Expect(lparenthesis, S1, lg1);          /* read left parenthesis */
    if (symb.ibsymb != rparenthesis) {

	/* parse parameter list and store descriptors in us_arg[] */

	paramList(i, us_fp, us_arg, us_arg_lg, S2, lg2);
    }
    Expect(rparenthesis, Stop, lg);         /* read right parenthesis */
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.paramList
 :*
 :* TITLE         (/ parsing the list of parameter descriptors excluding
 :*                  the parenthesis                                    /)
 :* PURPOSE       (/ parses a parameter descriptor list of a predefined
 :*                  routine and install their descriptors in us_arg[] if no
 :*                  error is found or recover from this error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-08
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - i          is an index of the routine descriptor in us_fp[]
 :*             - us_fp[]    is an array of predefined routine descriptors;
 :*             - us_arg[]   is an array of routine argument descriptors;
 :*             - us_arg_lg  is a pointer to the length of us_arg[]
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)  /)
 :*
 :*                                                             END-Pre
 :*
 :*   Post  (/ if
 :*              - the CSS if of the form:
 :*
 :*                 <paramList> <rest>
 :*
 :*                where  <paramList> is a valid parameter descriptor list
 :*                       <rest> is any character sequence
 :*              - us_fp[k].name <> routine_name (0<= k < *us_fp_lg)
 :*              - *us_fp_lg < Maxfp
 :*
 :*            then
 :*             1) if <paramList> is empty
 :*                then nothing is done
 :*             2) if <paramList> is of the form:
 :*                         "undef"<passing_mode>
 :*                  where <passing_mode> is "ref" or "val"
 :*                then
 :*                  - us_fp[i].arity = -1
 :*                  - us_fp[i].argp = ref   if <passing_mode> = "ref"
 :*                                  = value if <passing_mode> = "val"
 :*             3) if <paramList> is of the form: (n>=1)
 :*                   <type_1>":"<mode_1>"," ... ","<type_n>":"<mode_n>
 :*                  where <type_i> = "integer" or "string"
 :*                        <mode_i> = "ref" or "val"
 :*                  - us_fp[i].arity = n
 :*                  - for (j = 0, ..., n)
 :*                    us_arg[lg_0 + j].tpass = ref   if <mode_j> = "ref"
 :*                                           = value if <mode_j> = "val"
 :*                    us_arg[lg_0 + j].targ
 :*                                 = integer    if <type_j> = "integer"
 :*                                 = bytestring if <type_j> = "string"
 :*
 :*            in all cases, the CSS is <rest>
 :*          Otherwise
 :*              - the CSS is of the form:
 :*                     <first_Stop><rest>
 :*                where <first_Stop> is the first symbol in
 :*                             s1 s2 ... sN
 :*                      such that <first_Stop> in 'Stop[]'
 :*              - one or more error messages are output to stdout
 :*              - 'ErrorFound' is set to TRUE
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.                             /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.paramList
 :******************************************************************
*/
paramList(i, us_fp, us_arg, us_arg_lg, Stop, lg)
int i;                      /* index of the descr. in us_fp[]          */
fp_descr *us_fp;            /* table of predefined routine descriptors */
arg_descr *us_arg;          /* table of routine argument descriptors   */
int *us_arg_lg;             /* pointer to the length of us_arg[]       */
int Stop[], lg;             /* stops */
{
    int arity;              /* number of arguments */

    if (symb.ibsymb == undefx) {     /* is this a variable arity routine */
	variable_arity(i, us_fp, Stop, lg);
	if (i != -1)
	    us_fp[i].narg = undefArity;
    } else if (symb.ibsymb == intx || symb.ibsymb == strx) {

	/* parse parameter list and store descriptors in us_arg[] */

	arity = paramList1(us_arg, us_arg_lg, Stop, lg);
	if (i != -1) {
	    us_fp[i].narg = arity;  /* store the arity */
	}
    } else {
	error(syntax, 0);       /* a syntax error was found */
    }
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.variable_arity
 :*
 :* TITLE         (/ parsing the (common) descriptor of a varible arity
 :*                  predefined routine /)
 :* PURPOSE       (/ parses the common argument descriptor of a predefined
 :*                  routine and install it in us_arg[] if no error was
 :*                  found or recover from this error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-08
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - i          is an index of the routine descriptor in us_fp[]
 :*             - us_fp[]    is an array of predefined routine descriptors;
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)  /)
 :*
 :*                                                             END-Pre
 :*
 :*   Post  (/  if the CSS is of the form:
 :*                         "undef"<passing_mode>
 :*                  where <passing_mode> is "ref" or "val"
 :*                then
 :*                  - us_fp[i].arity = -1
 :*                  - us_fp[i].argp = ref   if <passing_mode> = "ref"
 :*                                  = value if <passing_mode> = "val"
 :*                  - the CSS is <rest>
 :*          Otherwise
 :*              - the CSS is of the form:
 :*                     <first_Stop><rest>
 :*                where <first_Stop> is the first symbol in
 :*                             s1 s2 ... sN
 :*                      such that <first_Stop> in 'Stop[]'
 :*              - one or more error messages are output to stdout
 :*              - 'ErrorFound' is set to TRUE
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.                             /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.variable_arity
 :******************************************************************
*/
variable_arity(i, us_fp, Stop, lg)
int i;              /* index in us_fp[] of the routine */
fp_descr *us_fp;    /* table of predefined routine descriptors */
int Stop[], lg;     /* the stops */
{
    int S1[Mlg], lg1, S2[Mlg], lg2, S3[Mlg], lg3; /* auxiliary stops */

   if (i != -1) {

/*** this test is disabled to permit variable argument routines to return
/*** a value 	Thu Oct  5 15:34:04 MET 1995 (AZIZ MOUNJI)
***/
	if (0 /* us_fp[i].rettype != undef */ /* fieldstring */) {
	    error(semantics, 4); /* a procedure does not return a value */
	}
    }

    /* build stops */
    insert_stop(valx, Stop, lg, S3, &lg3);
    insert_stop(refx, S3, lg3, S2, &lg2);
    insert_stop(colon, S2, lg2, S1, &lg1);
    Expect(undefx, S1, lg1);                  /* read "undef" */
    Expect(colon, S2, lg2);                   /* read colon   */
    if (symb.ibsymb == valx) {                /* all argument are by value */
	Expect(valx, Stop, lg);
	if (i != -1)
	    us_fp[i].argp = value;            /* store passing mode */
    } else if (symb.ibsymb == refx) {         /* all argument are by ref.  */
	Expect(refx, Stop, lg);
	if (i != -1)
	    us_fp[i].argp = ref;
    } else {
	error(syntax, 3);                     /* 'val' or 'ref' expected   */
    }
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.paramList1
 :*
 :* TITLE         (/ parsing the list of parameter descriptors for a fixed
 :*                  arity predefined routine                           /)
 :* PURPOSE       (/ parses a parameter descriptor list of a fixed arity
 :*                  predefined routine and install their descriptors in
 :*                  us_arg[] if no error is found or recover from this
 :*                  error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-08
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - us_arg[]   is an array of routine argument descriptors;
 :*             - us_arg_lg  is a pointer to the length of us_arg[]
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)  /)
 :*
 :*                                                             END-Pre
 :*
 :*   Post  (/  If the CSS is of the form: (n>=1)
 :*                <type_1>":"<mode_1>"," ... ","<type_n>":"<mode_n> <rest>
 :*                  where <type_i> = "integer" or "string"
 :*                        <mode_i> = "ref" or "val"
 :*             then
 :*                  - us_fp[i].arity = n
 :*                  - for (j = 0, ..., n)
 :*                    us_arg[lg_0 + j].tpass = ref   if <mode_j> = "ref"
 :*                                           = value if <mode_j> = "val"
 :*                    us_arg[lg_0 + j].targ
 :*                                 = integer    if <type_j> = "integer"
 :*                                 = bytestring if <type_j> = "string"
 :*
 :*                  - the CSS is <rest>
 :*          Otherwise
 :*              - the CSS is of the form:
 :*                     <first_Stop><rest>
 :*                where <first_Stop> is the first symbol in
 :*                             s1 s2 ... sN
 :*                      such that <first_Stop> in 'Stop[]'
 :*              - one or more error messages are output to stdout
 :*              - 'ErrorFound' is set to TRUE
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  "int" (/ the number of argument descriptors       /)
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.                             /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.paramList1
 :******************************************************************
*/
paramList1(us_arg, us_arg_lg, Stop, lg)
arg_descr *us_arg;    /* table of argument descriptors       */
int *us_arg_lg;       /* pointer to the length of this table */
int Stop[], lg;       /* the stops                           */
{
    int S1[Mlg], lg1, S2[Mlg], lg2, S3[Mlg], lg3;  /* auxiliary stops */
    int arity = 1;

    /* build stops */
    insert_stop(comma, Stop, lg, S1, &lg1);
    insert_stop(strx, S1, lg1, S2, &lg2);
    insert_stop(intx, S2, lg2, S3, &lg3);
    param(us_arg, us_arg_lg, S1, lg1);
    while (symb.ibsymb == comma) {         /* loop through the list        */
	Expect(comma, S3, lg3);            /* read the comma               */
	param(us_arg, us_arg_lg, S1, lg1); /* parse the next descriptor    */
	arity++;                           /* increment the number os arg. */
    }
    return (arity);
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.param
 :*
 :* TITLE         (/ parses a parameter descriptor                      /)
 :*                  arity predefined routine                           /)
 :* PURPOSE       (/ parses a parameter descriptor of a predefined routine
 :*                  and install it in us_arg[] if no error is found or
 :*                  recover from this error otherwise /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-08
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - us_arg[]   is an array of routine argument descriptors;
 :*             - us_arg_lg  is a pointer to the length of us_arg[]
 :*             - Stop[]     is an array of stops (see REMARKS below)
 :*             - lg         is the length of Stop[]
 :*             - the CSS    is s1 s2 ... sn      (see REMARKS below)  /)
 :*
 :*                                                             END-Pre
 :*
 :*   Post  (/  If the CSS is of the form: (n>=1)
 :*                <type>":"<mode><rest>
 :*                  where <type> = "integer" or "string"
 :*                        <mode> = "ref" or "val"
 :*                        <rest> is any character sequence
 :*             then
 :*                  -  us_arg[*us_arg_lg].targ
 :*                                 = integer    if <type_j> = "integer"
 :*                                 = bytestring if <type_j> = "string"
 :*
 :*                  -  us_arg[*us_arg_lg].tpass = ref   if <mode> = "ref"
 :*                                              = value if <mode> = "val"
 :*                  - the CSS is <rest>
 :*          Otherwise
 :*              - the CSS is of the form:
 :*                     <first_Stop><rest>
 :*                where <first_Stop> is the first symbol in
 :*                             s1 s2 ... sN
 :*                      such that <first_Stop> in 'Stop[]'
 :*              - one or more error messages are output to stdout
 :*              - 'ErrorFound' is set to TRUE
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - let the input stream be:
 :*                                 <t2><t3>...<tn>
 :*                     the symbol sequence:
 :*                             <t1><t2><t3>...<tn>
 :*                     where <t1> is the last symbol been read is called the
 :*                     'current symbol sequence' or CSS for short
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.                             /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.param
 :******************************************************************
*/
param(us_arg, us_arg_lg, Stop, lg)
arg_descr *us_arg;      /* table of argument descriptors       */
int *us_arg_lg;         /* pointer to the length of this table */
int Stop[], lg;         /* the stops */
{
    int S1[Mlg], lg1, S2[Mlg], lg2, S3[Mlg], lg3;  /* auxiliary stops */

    /* build stops */
    insert_stop(valx, Stop, lg, S3, &lg3);
    insert_stop(refx, S3, lg3, S2, &lg2);
    insert_stop(colon, S2, lg2, S1, &lg1);
    if (symb.ibsymb == intx) {               /* if the type is integer */
	Expect(intx, S1, lg1);
	us_arg[*us_arg_lg].targ = integer;   /* store the type         */
    } else if (symb.ibsymb == strx) {
	Expect(strx, S1, lg1);
	us_arg[*us_arg_lg].targ = bytestring;
    } else {
	error(syntax, 2);
    }
    Expect(colon, S2, lg2);                         /* read the colon     */
    if (symb.ibsymb == valx) {                      /* value argument     */
	Expect(valx, Stop, lg);
	us_arg[(*us_arg_lg)++].tpass = value;       /* store passing mode */
    } else if (symb.ibsymb == refx) {
	Expect(refx, Stop, lg);
	us_arg[(*us_arg_lg)++].tpass = ref;
    } else {
	error(syntax, 3);                     /* 'val' or 'ref' expected" */
    }
}
/*
 :******************************************************************
 :* UNIT-BODY parse.install_descr
 :*
 :* TITLE         (/ install a new argument descriptor   /)
 :*
 :* PURPOSE       (/ add a new descriptor to the table us_fp[] such that
 :*                  the table remains sorted on routine names /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-07-08
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                                                               /)
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - 'name' is a string representing a routine name
 :*             - the set of entries in 'us_fp[]' is:
 :*               S = {(name_0, ...), ..., (name_n, ...)}  (n=*us_fp_lg-1)
 :*             - 'us_fp[]' is sorted in ascending alphabetic order
 :*               on routine names:
 :*      us_fp[0].name < ... < us_fp[*us_fp_lg-1].name
 :*             - '*us_fp_lg' = lg_0
 :*                                                                     /)
 :*                                                             END-Pre
 :*
 :*   Post  (/ - if '*us_fp_lg'<'MaxFp'-1 and 'name' not in S:
 :*                 - '*us_fp_lg' = lg_0 + 1
 :*                 - the set of entries in 'us_fp[]' is:
 :*                              S U {(name, ...)}
 :*                 - 'us_fp[]' is sorted in ascending alphabetic
 :*                   order on routine names:
 :*                       us_fp[0].name < ...
 :*                       < us_fp[*us_fp_lg-1].name
 :*                 - the returned value is 'i' such that:
 :*                       us_fp[i].name = name
 :*            - if '*us_fp_lg'<'MaxFp'-1 and 'name' in S:
 :*                 the following error message:
 :*                 "already part of the library: name ",
 :*                 is output on stdout
 :*            - otherwise if '*us_fp_lg'>='MaxFp'-1 the error message:
 :*                       "table overflow!: tfp_descr"
 :*               is output on stdout
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  "int"   (/ index of the installed argument descriptor /)
 :*
 :*   REMARKS          (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parse.install_descr
 :******************************************************************
*/
install_descr(us_fp, us_fp_lg, us_arg_lg, name)
fp_descr *us_fp;            /* table of routine descriptors */
int *us_fp_lg, *us_arg_lg;  /* pointers to table lengths    */
char *name;                 /* routine name                 */
{
    int i;  /* loop counter */
    int k;  /* to compare two strings */

    i = ++(*us_fp_lg);
    if (i >= MaxFp) {   /* check for table overflow */
	error1(fatal, 11, "tfp_descr");
	exit(1);
    }
    /* search for a place to install the descriptor */
    while (i-- > 1) {
	if ((k = strcmp(us_fp[i-1].name, name)) > 0) {
	    us_fp[i] = us_fp[i-1];   /* shift to the right */
	} else if (k < 0) {          /* the place is found */
	    break;
	} else { /* k = 0 */
	    error1(semantics, 5, name);  /* duplicate name */
	    exit(1);
	}
    }
    strcpy(us_fp[i].name, name);         /* copy the name */
    us_fp[i].narg = 0;          /* store the index to this entry */
    us_fp[i].rettype = undef;          /* store the index to this entry */
    us_fp[i].argp = *us_arg_lg;          /* store the index to this entry */
    return(i);                           /* return index */
}
/*
 :***********************************************************************
 :* UNIT-BODY parseLib.linkLib
 :*
 :* TITLE         (/ generate a make file for Asax   /)
 :*
 :* PURPOSE       (/ generate a tex file which is an executable file that
 :*                  recompile and relink some ASAX and user modules in order
 :*                  to generate a new ASAX executable               /)
 :*
 :*
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-06
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                     - tObjMod[] : an entry in this table
 :*                                            represents the name of the
 :*                                            of the object module
 :*                                            implementing a predefined
 :*                                            routine
 :*                     - tModObj_lg         : length of 'tModObj[]'
 :*                     - MaxObjMod          : maximum length of 'tModObj[]'
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator
 :*                               ASAX: Specification of
 :*                          some extensions to the analyzer           /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - tObjMod[] is an array of 'tObjMod_lg' module object names;
 :*             - makeFile  is a string repreesenting the name of the file
 :*                         to be generated;
 :*                                                                      /)
 :*                                                             END-Pre
 :*   Post   (/
 :*          CONTENT:
 :*           * The make file: (mode 777)
 :*                Let module_1, ..., module_k (k>=0) be the names of the
 :*                object modules listed in the two input files. Then the
 :*                make file (makeAsax or 'prefix'.makeAsax) have the
 :*                following content:
 :*
 :*                lib_files='module_1.o module2.o ...  module_k.o'
 :*                obj_file=obj_file_name
 :*                mode_file=standard.o
 :*                asaxExec=asax.exec
 :*
 :*
 :*                cc -c std.c
 :*
 :*                case $1 in
 :*                        0) obj=NADF_file.o;;
 :*                        *) obj="$obj_file NADF_file.o"
 :*                esac
 :*
 :*
 :*                cc $mode_file emulate.o $obj  \
 :*                stack.o parse.o usage.o globalVar.o rule_decl.o action.o \
 :*                guardAct.o condition.o trigger.o arglist.o term.o  \
 :*                expr.o string.o printbox.o std.o recov.o rdescr.o \
 :*                error.o merge.o $lib_files scan.o  -o $asaxExec
 :*
 :*                                                                       /)
 :*                                                             END-Post
 :*   RETURNS NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*
 :*   REMARKS          (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parseLib.linkLib
 :******************************************************************
*/
linkLib(makeFile)
char *makeFile;   /* the name of the file to be generated */
{
    int i;          /* loop counter */
    char *prefix;   /* object module name without .o extension */
    char *strcat();
    FILE *fd;       /* file descriptor of the generated file */

    if ((prefix = (char *) malloc(2000)) == NULL) {
	error(fatal, 9);           /* allocation problem */
	exit(1);
    }
    strcpy(prefix, "");

     /* build a string equal to the list of object module names */

    for (i=tObjSysMod_lg; i<tObjMod_lg; i++) {
	 prefix = strcat(prefix, tObjMod[i]);
	 prefix = strcat(prefix, ".o ");
    }
    if ((fd = fopen(makeFile, "w")) == NULL) { /* open file for writing */
	error1(fatal, 6, makeFile);         /* error during open */
	exit(1);
    }
	 /* output text to the text file */

    fprintf(fd, "%s\'%s\'\n", apath, ASAXPATH);
    fprintf(fd, "%s\'%s\'\n", lib_files, prefix);
    fprintf(fd, "%s\n%s\n%s\n\n", obj_file, mode_file, asaxExec);
    fprintf(fd, "%s -c -I. $APATH/std.c -o .__std.o\n\n", COMPILER);
    fprintf(fd, "%s %s\n\n", COMPILER, cc1);
    fclose(fd);             /* close the file */
}
/*
 :***********************************************************************
 :* UNIT-BODY parseLib.dparse_LDF
 :*
 :* TITLE         (/ generate a C-initialization file and prototype file /)
 :*
 :* PURPOSE       (/ generate C-initialization file and prototypes file
 :*                  given the 4 global tables.                 /)
 :*
 :*
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-06
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA       (/
 :*                - sys_fp[]     : table of predefined routine descriptors
 :*                - sys_fp_lg    : length of sys_fp[]
 :*                - user_fp[]    : table of user defined routine descriptors
 :*                - user_fp_lg   : length of user_fp[]
 :*                - tfp_descr[]  : table of all defined routine descriptors
 :*                                 (user defined + predefined)
 :*                - tfp_descr_lg : length of tfp_descr[]
 :*                - targ_descr[] : table of all argument descriptors
 :*                                 (user defined + predefined)
 :*                - targ_descr_lg: length of targ_descr[]
 :*                - MaxFp        : max. length of routine descriptor tables
 :*                - MaxFpArg     : max. length of arg. descriptor tables
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator
 :*
 :*                               ASAX: Specification of
 :*                          some extensions to the analyzer           /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - 'fd1' a valid file descriptor for the prototype file
 :*             - 'fd2' a valid file descriptor for the C-initialization file
 :*
 :*                                                                      /)
 :*                                                             END-Pre
 :*   Post   (/
 :*
 :*            * the file corresponding to fd1 contains the C-prototypes
 :*              of the C-routines listed in 'sys_fp[]' U 'user_fp[]'
 :*
 :*            * the file corresponding to fd contains the C-initialization
 :*              of the table tfp_descr[] ( = sys_fp[] U user_fp[]) and the
 :*              C-initialization of targ_descr[] (= sys_arg[] U user_arg[])
 :*                                                                       /)
 :*                                                             END-Post
 :*   RETURNS NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*
 :*   REMARKS          (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parseLib.dparse_LDF
 :******************************************************************
*/
dparse_Ldf(fd1, fd2)
FILE *fd1;   /* file descriptor for prototype file            */
FILE *fd2;   /* file descriptor for the C-initialization file */
{
    int i;                        /* loop counter                        */
    char c1 = ',';                /* comma for prototype declaration     */
    static char s0[2]={'"','\0'}; /* hard to put directly in the code!   */
    char *s;                      /* auxiliary string                    */
    char *s1;                     /* contains the routine name in quotes */
    

	/* allocate area for routine names */
    if ((s = s1 = (char *) malloc(ident_lg)) == NULL) {
	error(fatal, 9);           /* allocation problem */
	exit(1);
    }
    *s++ = '"';
    mergeLib(); /* merge sys_fp[]  and user_fp[]  in tfp_descr[]  */
		/* merge sys_arg[] and user_arg[] in targ_descr[] */

    fprintf(fd2, "#define tfp_descr_lg %d\n\n", tfp_descr_lg);
    fprintf(fd2, "fp_descr tfp_descr[tfp_descr_lg] = {\n");
    for (i=0; i<tfp_descr_lg; i++) {

		 /* generate prototypes */

	if (!i)
	    fprintf(fd1, "\nvoid *%s()", tfp_descr[i].name);
	else if (!(i%4))   /* 4 prototypes  per line */
	    fprintf(fd1, ";\nvoid *%s()", tfp_descr[i].name);
	else
	    fprintf(fd1, "%c *%s()", c1, tfp_descr[i].name);
	if (i)             /* not a new line of prototypes */
	    fprintf(fd2, "%c\n", c1);
	strcpy(s, tfp_descr[i].name);
	strcpy(s+strlen(tfp_descr[i].name), s0);

	/* C-initialization of tfp_descr[] */


	fprintf(fd2, "  %-16s\t,%-16s\t,%d\t,%-10s\t,%d",
		 s1,
		tfp_descr[i].name,
		tfp_descr[i].narg,
		typesTab[(int) (tfp_descr[i].rettype)],
		tfp_descr[i].argp);
    }
    fprintf(fd1, ";\n");
    fprintf(fd2, "\n\t\t};\n\n");

	/* C-initialization of targ_descr[] */

    fprintf(fd2, "#define targ_descr_lg %d\n\n", targ_descr_lg);
    fprintf(fd2, "arg_descr targ_descr[targ_descr_lg] = {\n");
    for (i=0; i<targ_descr_lg; i++) {
	if (i) fprintf(fd2, "%c\n", c1);
	fprintf(fd2, "\t\t%s\t,%s",
		passtypeTab[(int) (targ_descr[i].tpass)],
		typesTab[(int) (targ_descr[i].targ)]);
    }
    fprintf(fd2, "\n\t\t};\n\n");
}
/*
 :******************************************************************
 :* UNIT-BODY parseLib.mergeLib
 :*
 :* TITLE         (/ merging of descriptor tables related to the library /)
 :*
 :* PURPOSE       (/ merges sys_fp[]  and user_fp[]  in tfp_descr[]
 :*                  merges sys_arg[] and user_arg[] in targ_descr[] /)
 :*
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1993-07-08
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA  (/
 :*                                                                  /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*
 :*                               ASAX: Specification of
 :*                          some extensions to the analyzer           /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre (/ - 'sys_fp' is an array of routine descriptors sorted in
 :*            ascending alphabetic order on the field 'name'
 :*          - 'sys_fp_lg' is the length of 'sys_fp[]'
 :*          - {s_1, ..., s_sys_fp_lg} (='S') is the set of descriptors in
 :*            'sys_fp[]'
 :*          - 'user_fp' is an array of routine descriptors sorted in
 :*            ascending alphabetic order on the field 'name'
 :*          - 'user_fp_lg' is the length of 'user_fp[]'
 :*          - {u_1, ..., u_user_fp_lg} (='U') is the set of descriptors in
 :*            'user_fp[]'
 :*                                                             END-Pre
 :*
 :*   Post (/ - 'tfp_descr[]' is an array of routine descriptors sorted in
 :*             ascending alphabetic order on the field 'name'
 :*           - 'S' U 'U' is the set of routine descriptors in 'tfp_descr[]'
 :*           - tfp_descr_lg = sys_fp_lg + user_fp_lg
 :*           - for (i=sys_fp_lg, ..., sys_fp_lg + user_fp_lg - 1)
 :*             if user_fp[i - sys_fp_lg].argp <> -1
 :*              tfp_descr[i].argp = user_fp[i - sys_fp_lg].argp + sys_arg_lg
 :*             else
 :*                 tfp_descr[i].argp = user_fp[i - sys_fp_lg].argp
 :*           - for (i=0, ... sys_arg_lg - 1)
 :*                - targ_descr[i] = sys_arg[i]
 :*           - for (i=sys_arg_lg, ... sys_arg_lg + user_arg_lg - 1)
 :*                - targ_descr[i] = user_arg[i - sys_arg_lg]
 :*           - targ_descr_lg = sys_arg_lg + user_arg_lg
 :*                                                              /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS     (/ Definitions:
 :*                   - the values of the enumerated type 'enum index' are
 :*                     called 'stops'. They are used as indexes to the table
 :*                     'specsymbtab[]' so as a 'stop' identifies an element
 :*                     this table.
 :*                                                         /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY parseLib.mergeLib
 :******************************************************************
*/
mergeLib()
{
    int i, j, k;  /* indexes in sys_fp[], user_fp and tfp_descr[] respec. */
    int l;        /* result of string comparison */

	    /* initialize counters */
    i = 0;
    j = 0;
    k = 0;
	    /* merge sys_fp[] and user_fp[] in tfp_descr[] */

    while (i< sys_fp_lg && j< user_fp_lg && k < MaxFp) {
	if ((l = strcmp(sys_fp[i].name, user_fp[j].name)) < 0) {
	    tfp_descr[k++] = sys_fp[i++];
	} else if (l > 0) {
/* 
 *	Aziz Mounji: Wed May 29 10:17:41 WET DST 1996 
 *
 * 	Must be done only in case this is not a variable arity routine
 * 	in which case argp is value or ref and is not an index in 
 *	targ_descr[]. The next test is added to see if we 're dealing
 * 	with a variable arity routine.
 */
	    if (user_fp[j].narg != undefArity) 
	    	user_fp[j].argp += sys_arg_lg;
	    tfp_descr[k++] = user_fp[j++];
	} else {
	    error1(semantics, 5, user_fp[j].name);  /* duplicate name */
	    exit(1);
	}
    }
       /* check for table overflow */

    if (k == MaxFp) {
	error1(fatal, 8, "tfp_descr");
	exit(1);
    }
    if (j == user_fp_lg) {  /* end of user_fp[] reached first */
	for (; i< sys_fp_lg; i++, k++) {
	    tfp_descr[k] = sys_fp[i];  /* add remaining descriptors */
	}
    } else {                /* end of sys_fp[] reached first */
	for (; j< user_fp_lg; j++, k++) {
	    if (user_fp[j].argp != -1) /* adjust indexes in tfp_descr[] */
		user_fp[j].argp += sys_arg_lg;
	    tfp_descr[k] = user_fp[j];     /* add remaining descriptors     */
	}
    }
    tfp_descr_lg = user_fp_lg + sys_fp_lg;

	    /* merge sys_arg[] and user_arg[] in targ_descr[] */

    i=0;
    while (i<sys_arg_lg && i<MaxFpArg) {
	targ_descr[i] = sys_arg[i];
	i++;
    }
    /* check for table overflow */
    if (i == MaxFpArg) {
	error1(fatal, 8, "targ_descr");
	exit(1);
    }
    i=0;
    while (i<user_arg_lg && i<MaxFpArg - sys_arg_lg) {
	targ_descr[sys_arg_lg + i] = user_arg[i];
	i++;
    }
    if (i == MaxFpArg - sys_arg_lg) {
	error1(fatal, 8, "targ_descr");
	exit(1);
    }
    targ_descr_lg = sys_arg_lg + user_arg_lg;
}

